// SQL14.cpp : main project file.

using namespace System;
using namespace System::Diagnostics;
using namespace System::Threading;

/* Raima::Rdm is the system module that contains the ADO.NET
 * data provider for RDM */
using namespace Raima::Rdm;

/* EXAMPLE - SQL14 for C++ using ADO.NET */
/* This example shows multi-user access to a database using separate processes
 * and a remote tfserver.  The connections will use TCP/IP to communicate
 * with the tfserver to do their work.  After starting the tfserver, two
 * writer processes are created.  The first begins inserting values into the
 * table.  The second waits until the table contains more than 100 rows and
 * then begins deleting the oldest.  A reader process is also created that
 * prints out the current oldest and newest entries every second.  The reader
 * process will use read-only transactions.
 * 
 * The following methods are illustrated:
 *     Raima::Rdm::RdmConnection:
 *         BeginTransaction() Method
 *         ChangeDatabase(string) Method
 *         Close() Method
 *         CreateCommand() Method
 *         Open() Method
 *         RdmConnection() Constructor
 *     Raima::Rdm::RdmCommand:
 *         CommandText Property
 *         ExecuteNonQuery() Method
 *         ExecuteReader() Method
 *         Parameters Property
 *     Raima::Rdm::RdmDataReader:
 *         Item[int] Property
 *         Read() Method
 *     Raima.Rdm.RdmParameterCollection
 *         Add(string, RdmType)
 *         Item[string] Property
 * 
 * They are fully document in http://docs.raima.com/rdm/.
 * 
 * For simplicity this example does not catch any thrown execptions
 * but good programming practices would dictate that this occur in
 * a full application.
 */

static void DropDatabase(
    RdmConnection ^conn,
    String        ^name)
{
    RdmCommand ^cmd = conn->CreateCommand();

    /* Attempt to drop this database.  If the database does not already exist
     * this will fail and we simply ignore the failure for this sample. */
    try
    {
        cmd->CommandText = "DROP DATABASE " + name;
        cmd->ExecuteNonQuery();

        cmd->CommandText = "COMMIT";
        cmd->ExecuteNonQuery();
    }
    catch (Exception ^)
    {
    }
}

static void CreateThisDatabase(
    RdmConnection   ^conn,
    String          ^name,
    array<String ^> ^cmds)
{
    RdmCommand ^cmd = conn->CreateCommand();

    cmd->CommandText = "CREATE DATABASE " + name;
    cmd->ExecuteNonQuery();

    for (int ii = 0; ii < cmds->Length; ii++)
    {
        cmd->CommandText = cmds[ii];
        cmd->ExecuteNonQuery();
    }

    cmd->CommandText = "COMMIT";
    cmd->ExecuteNonQuery();

    cmd->CommandText = "CLOSE " + name;
    cmd->ExecuteNonQuery();
}

static void CreateDatabases(
    RdmConnection ^conn)
{
    array<String ^> ^cmds = {
        "CREATE TABLE Table1 (id INTEGER PRIMARY KEY)"
    };

    /* Drop any existing copy of the databases. */
    DropDatabase(conn, "SQL14");

    /* Attempt to create the databases */
    CreateThisDatabase(conn, "SQL14", cmds);
}

static void Inserter()
{
    Console::WriteLine("Inserter process");

    /* Create a connection for this process */
    RdmConnection ^conn = gcnew RdmConnection("host=local;database=SQL14;openmode=s;tfstype=tfsr");
    conn->Open();

    /* Create a command to do the insert */
    RdmCommand ^cmd = conn->CreateCommand();
    cmd->CommandText = "INSERT INTO Table1 VALUES :id";
    cmd->Parameters->Add("id", RdmType::Int32);

    /* Loop until a shutdown is requested */
    for (int ii = 0; ; ii++)
    {
        /* Insert the next value */
        cmd->Parameters["id"]->Value = ii;
        cmd->ExecuteNonQuery();
    }
}

static void Deleter()
{
    Console::WriteLine("Deleter process");

    /* Create a connection for this process */
    RdmConnection ^conn = gcnew RdmConnection("host=local;database=SQL14;openmode=s;tfstype=tfsr");
    conn->Open();

    /* Create a command to check the number of rows, one to fine the 'smallest'
     * row, and one to delete */
    RdmCommand ^chkCmd = conn->CreateCommand();
    RdmCommand ^minCmd = conn->CreateCommand();
    RdmCommand ^delCmd = conn->CreateCommand();

    chkCmd->CommandText = "SELECT COUNT(id) FROM Table1";
    minCmd->CommandText = "SELECT MIN(id) FROM Table1";
    delCmd->CommandText = "DELETE FROM Table1 WHERE id = :minid";
    RdmParameter ^parm = delCmd->Parameters->Add("minid", RdmType::Int32);

    /* Loop until a shutdown is requested */
    while (true)
    {
        /* Query the number of rows in the table */
        RdmDataReader ^countRdr = chkCmd->ExecuteReader();
        if (countRdr->Read())
        {
            long count = (long) (Int64) countRdr[0];

            /* There should be only one row, so this closes the reader */
            countRdr->Read();

            if (count > 100)
            {
                /* If there at least 100 rows, get the smallest and delete it */
                RdmDataReader ^minRdr = minCmd->ExecuteReader();
                if (minRdr->Read())
                {
                    parm->Value = (int) minRdr[0];
                    delCmd->ExecuteNonQuery();

                    /* There should be only one row, so this closes the reader */
                    minRdr->Read();
                }
            }
        }
    }
}

static void Reader()
{
    Console::WriteLine("Reader process (using read-only transactions)");

    /* Create a connection for this process */
    RdmConnection ^conn = gcnew RdmConnection("host=local;database=SQL14;openmode=s;tfstype=tfsr;readonlytransactions=1");
    conn->Open();

    /* Create a command to do the read */
    RdmCommand ^cmd = conn->CreateCommand();
    cmd->CommandText = "SELECT MIN(id), MAX(id) FROM Table1";

    /* Loop until a shutdown is requested */
    while (true)
    {
        /* Wait one second */
        Thread::Sleep(1000);

        /* Query the mininum and maximum values */
        RdmDataReader ^rdr = cmd->ExecuteReader();
        if (rdr->Read())
        {
            Console::WriteLine("Current minimum = " + rdr[0] + " and current maximum = " + rdr[1]);
            rdr->Read();
        }
    }
}

int main(array<System::String ^> ^args)
{
    if (args->Length > 0)
    {
        if (args[0] == "inserter")
            Inserter();
        else if (args[0] == "deleter")
            Deleter();
        else if (args[0] == "reader")
            Reader();
    }
    else
    {
        Console::WriteLine("Starting tfserver (and waiting 5 seconds for it to initialize)");

        /* Start the tfserver process to process 'remote' connections */
        Process ^tfserver = gcnew Process();
        tfserver->StartInfo->FileName = "Debug\\tfserver.exe";
        tfserver->Start();

        /* Sleep to give the tfserver time to initialize and start listening */
        Thread::Sleep(5000);

        /* Create the connection with a simple connection string */
        RdmConnection ^conn = gcnew RdmConnection("host=local;openmode=s;tfstype=tfsr");
        conn->Open();

        CreateDatabases(conn);

        /* Create three processes one to insert, one to delete, and one to show the
         * current contents */
        Process ^inserter = gcnew Process();
        inserter->StartInfo->FileName = "Debug\\SQL14_CPP_ADO.exe";
        inserter->StartInfo->Arguments = "inserter";

        Process ^deleter = gcnew Process();
        deleter->StartInfo->FileName = "Debug\\SQL14_CPP_ADO.exe";
        deleter->StartInfo->Arguments = "deleter";

        Process ^reader = gcnew Process();
        reader->StartInfo->FileName = "Debug\\SQL14_CPP_ADO.exe";
        reader->StartInfo->Arguments = "reader";

        Console::WriteLine("Starting processes");
        Console::WriteLine("   Press <Enter> when you are ready to terminate them");

        /* Start each of the processes */
        inserter->Start();
        deleter->Start();
        reader->Start();

        /* Wait for user to hit <Enter> */
        Console::ReadLine();

        /* Shutdown the processes */
        inserter->CloseMainWindow();
        deleter->CloseMainWindow();
        reader->CloseMainWindow();

        /* Wait for each process to finish */
        inserter->WaitForExit();
        deleter->WaitForExit();
        reader->WaitForExit();

        Console::WriteLine("All processes have terminated");

        conn->Close();

        /* Stop the tfserver */
        tfserver->CloseMainWindow();
        tfserver->WaitForExit();
        tfserver->Close();

        Console::WriteLine("Hit <Enter> to continue");
        Console::ReadLine();
    }

    return 0;
}
